package statechart

import scala.collection.mutable.Map
import java.util.{Timer,TimerTask}

// All classes making up the statechart scala object representation.

class Statechart(n: String, ps: Map[String,Any], is: String, ss: List[State]) {
  val name = n
  val context = ps
  val initialState = is
  val states = ss  
  var currentState: State = null
  
  def start() = {
    currentState = getState(initialState)
    for (s <- states) {
      s.init(this)
    }
  }
  
  def trigger(event: String) {
    currentState = currentState.trigger(event)
  }  
  
  def getState(name: String): State = {
    for (s <- states)
      if (s.name == name)
        return s
    throw new RuntimeException("state unknown: " + name)
  }
  
  override def toString = "Statechart " + name + 
    "\n\tcurrent state: " + currentState + 
    "\n\tcontext: " + context
}

class State(n: String, oe: Option[List[String]], ox: Option[List[String]], ots: List[Transition]) {
  val name = n
  val outgoingTransitions = ots
  val onEntryEvent = oe match {
    case Some(a) => a
    case _ => null
  }  
  val onExitEvent = ox match {
    case Some(a) => a
    case _ => null
  }  
  var statechart: Statechart = null
  
  def init(s: Statechart) {
    statechart = s
  }
  
  def onEntry() {
    println(name + " onEntry")
    if (onEntryEvent != null) {
        for (e <- onEntryEvent) {
          doAction(e)
        }    
    }
  }
  
  def onExit() {
    println(name + " onExit")
    if (onExitEvent != null) {
      for (e <- onExitEvent) {
        doAction(e)
      }
    }
  }
  
  // die Actions sind leider hier hart kodiert
  def doAction(e: String) {
    println("doing action " + e)
    e match {
      case "powerOnMagnetron" =>  statechart.context("magnetron") = true
      case "startTimer" => startTimer
      case "lightOn" => statechart.context("light") = true
      case "powerOffMagnetron" => statechart.context("magnetron") = false
      case "stopTimer" => stopTimer
      case "lightOff" => statechart.context("light") = false
    }
  }
  
  def startTimer {
    val c = statechart.context("cookingTime").asInstanceOf[Int]
    if (c == 0) {
      println("beep!")
      statechart.trigger("timerExpired")
      return
    }
    statechart.context("cookingTime") = c - 1
    new Timer().schedule(new TimerTask() {
      def run {
        println("tick")
        startTimer
      }
    }, 1000);
  }
  
  def stopTimer {
  }
  
  def trigger(event: String): State = {
    for (t <- outgoingTransitions) {
      if (t.event == event) {
        if (t.guard == null || 
            t.guard.eval(statechart.context(t.guard.property))) {
          println(event + " triggered transition to " + t.nextState)
          return statechart.getState(t.nextState)
        }
      }
    }
    this
  }
  
  override def toString = "State(" + name + ")"
}

class ExtendedState(id: String, ss: List[String], is: String, outgoingTransitions: List[Transition])
  extends State(id, null, null, outgoingTransitions) {
  
  val substates = ss
  val initialState = is
  var currentState: State = null
  
  override def init(s: Statechart) {
    super.init(s)
    currentState = statechart.getState(initialState)
  }
  
  override def trigger(event: String): State = {
    var newState = super.trigger(event)
    if (newState != this) {
      println(event + " triggered transition to " + newState)
      return newState
    }
    newState = currentState.trigger(event)
    if (newState != currentState) {
      currentState.onExit
      newState.onEntry
      currentState = newState      
    }
    this
  }
  
  override def onEntry {
    println(name + " onEntry")
    currentState = statechart.getState(initialState)
  }
  
  override def onExit {
    println(name + " onExit")
    currentState = null
  }  
  
  override def toString = "ExtendedState(" + name + ", currentState=" + currentState + ")"
}

class Transition(e: String, g: Option[Condition], n: String) {
  val event = e
  val nextState = n  
  
  val guard = g match {
    case Some(a) => a
    case None => null
  }  
  
  override def toString: String = {
    var s = event
    if (guard != null) s += "[" + guard + "]"
    s += " -> " + nextState
    s
  }
}

class Condition(n: Option[_], ident: String, c: Option[Comparison]) {  
  
  val property: String = ident
  
  val not: Boolean = n match {
    case a: Some[_] => true
    case None => false
  }
  
  val comp: Comparison = c match {
    case Some(a) => a
    case None => null
  }

  def eval(propertyValue: Any): Boolean = {
    println("evaluating ["  + this + "] with value=" + propertyValue)
    propertyValue match {
      case b: Boolean => {
        if (not) !b else b
      }
      case i: Int => {    
        comp.operator match {
          case "eq" => i == comp.operand
          case "gt" => i > comp.operand
          case "lt" => i < comp.operand
        }
      }
      case _ => true
    }
  }
  
  override def toString = {
    var res = ""
    if (not) res += "!"
    res += property
    if (comp != null)
      res += " " + comp
    res
  }
}

class Comparison(op: String, numericLit: String) {
  val operator = op
  val operand = numericLit.toInt
  override def toString = operator + " " + operand
}
